@@ -28,6 +28,10 @@
 #include <X11/XKBlib.h>
 #endif
 
+#ifdef USE_XINPUT
+#include <X11/extensions/XInput.h>
+#endif
+
 #include <gtk/gtk.h>
 #include <gdk/gdk.h>
 #include <gdk/gdkx.h>
@@ -104,6 +108,10 @@ GtkWidget *mouse_tray_icon = NULL;
 GtkWidget *mouse_tray_image = NULL;
 #endif
 
+#ifdef USE_XINPUT
+static gboolean xinputpresent = FALSE;
+#endif
+
 static void get_mouse_values(int *accel_return, int *denom_return, int *thresh_return)
 {
     gdk_flush();
@@ -124,6 +132,115 @@ static void get_mouse_values(int *accel_return, int *denom_return, int *thresh_r
     gdk_error_trap_pop();
 }
 
+#ifdef USE_XINPUT
+static gboolean
+check_xinput_support (void)
+{
+    gint op_code, event, error;
+
+    return XQueryExtension (GDK_DISPLAY (), "XInputExtension", &op_code, &event, &error);
+}
+
+static gboolean
+xinput_device_has_buttons (XDeviceInfo *device_info)
+{
+    int i;
+    XAnyClassInfo *class_info;
+
+    class_info = device_info->inputclassinfo;
+    for (i = 0; i < device_info->num_classes; i++)
+    {
+        if (class_info->class == ButtonClass)
+        {
+            XButtonInfo *button_info;
+
+            button_info = (XButtonInfo *) class_info;
+            if (button_info->num_buttons > 0)
+            {
+                return TRUE;
+            }
+        }
+
+        class_info = (XAnyClassInfo *) (((guchar *) class_info) + class_info->length);
+    }
+
+    return FALSE;
+}
+
+static void
+set_xinput_mouse_values (gboolean right_handed)
+{
+    XDeviceInfo *device_info;
+    unsigned char *buttons;
+    gint n_buttons, n_devices, i;
+    gint idx_1 = 0, idx_3 = 1;
+    gsize buttons_capacity = 16;
+
+    device_info = XListInputDevices (GDK_DISPLAY (), &n_devices);
+
+    if (n_devices > 0)
+    {
+        buttons = g_new (guchar, buttons_capacity);
+    }
+    else
+    {
+        buttons = NULL;
+    }
+
+    for (i = 0; i < n_devices; i++)
+    {
+        XDevice *device = NULL;
+
+        if ((device_info[i].use != IsXExtensionDevice) || (!xinput_device_has_buttons (&device_info[i])))
+        {
+            continue;
+        }
+
+        gdk_error_trap_push ();
+        device = XOpenDevice (GDK_DISPLAY (), device_info[i].id);
+
+        if ((gdk_error_trap_pop () != 0) || (device == NULL))
+        {
+            continue;
+        }
+
+        n_buttons = XGetDeviceButtonMapping (GDK_DISPLAY (), device, buttons, DEFAULT_PTR_MAP_SIZE);
+        if (n_buttons > DEFAULT_PTR_MAP_SIZE)
+        {
+            buttons = g_alloca (n_buttons);
+            n_buttons = XGetDeviceButtonMapping (GDK_DISPLAY (), device, buttons, n_buttons);
+        }
+
+        for (i = 0; i < n_buttons; i++)
+        {
+            if (buttons[i] == 1)
+            {
+                idx_1 = i;
+            }
+            else if (buttons[i] == ((n_buttons < 3) ? 2 : 3))
+            {
+                idx_3 = i;
+            }
+        }
+
+        if ((!right_handed && (idx_1 < idx_3)) || (right_handed && (idx_1 > idx_3)))
+        {
+            buttons[idx_1] = ((n_buttons < 3) ? 2 : 3);
+            buttons[idx_3] = 1;
+        }
+
+        XSetDeviceButtonMapping (GDK_DISPLAY (), device, buttons, n_buttons);
+        XCloseDevice (GDK_DISPLAY (), device);
+    }
+    g_free (buttons);
+
+    if (device_info != NULL)
+    {
+        XFreeDeviceList (device_info);
+    }
+}
+#endif
+
 static void set_mouse_values(gboolean right_handed, int accel, int threshold)
 {
     unsigned char *buttons;
@@ -133,6 +250,13 @@ static void set_mouse_values(gboolean right_handed, int accel, int threshold)
     gdk_flush();
     gdk_error_trap_push();
 
+#ifdef USE_XINPUT
+    if (xinputpresent)
+    {
+        set_xinput_mouse_values (right_handed);
+    }
+#endif
+
     XChangePointerControl(GDK_DISPLAY(), True, True, accel, denominator, threshold);
 
     buttons = g_alloca (DEFAULT_PTR_MAP_SIZE);
@@ -786,6 +910,25 @@ create_channel(McsPlugin *mcs_plugin)
         mcs_manager_set_int(mcs_plugin->manager, "Net/DndDragThreshold", CHANNEL1, dnd_threshold);
     }
 
+#ifdef USE_XINPUT
+#ifdef DEBUG
+    g_message ("Querying XINPUT extension");
+#endif
+    if (check_xinput_support())
+    {
+#ifdef DEBUG
+        g_message ("XINPUT extension found");
+#endif
+        xinputpresent = TRUE;
+    }
+    else
+    {
+#ifdef DEBUG
+        g_message ("Your X server does not support XINPUT extension");
+#endif
+        xinputpresent = FALSE;
+    }
+#endif
     setting = mcs_manager_setting_lookup(mcs_plugin->manager, "Mouse/RightHanded", CHANNEL2);
     if(setting)
     {
